-- declare spell names and IDs
local vf = GetSpellInfo(194249)
local stm = GetSpellInfo(193223)
local pi = GetSpellInfo(10060)
local disp = GetSpellInfo(47585)
local sa = 148859
local swp = 589
local mb = 8092
local vb = 205448
local swd = 32379
local swv = 205351
local sc = 205385

-- declare bar related stuff
local mainFrame = CreateFrame("Frame", nil, UIParent)
local bgFrame = CreateFrame("Frame", "InsanityBGFrame", mainFrame)
local barFrame = CreateFrame("Frame", "InsanityBarFrame", bgFrame)
local drainFrame = CreateFrame("Frame", "InsanityDrainFrame", barFrame)
local capFrame = CreateFrame("Frame", "InsanityCapFrame", bgFrame)
local threshFrame = CreateFrame("Frame", "InsanityThreshFrame", bgFrame)
local border = 3
local barWidth = 500
local barHeight = 20

-- declare settings panel stuff
local addon = {}
addon.panel = CreateFrame("Frame", "InsanityBarPanel", UIParent)
addon.panel.name = "InsanityBar"
addon.loaded = false
local showTimesChBox = CreateFrame("CheckButton", "InsanityBarShowTimes", addon.panel, "UICheckButtonTemplate")
local showAlwaysChBox = CreateFrame("CheckButton", "InsanityBarShowAlways", addon.panel, "UICheckButtonTemplate")
local barScaleSldr = CreateFrame("Slider", "InsanityBarScale", addon.panel, "OptionsSliderTemplate")

-- declare variables which actually change during the addon process
local spells = {}
local extraInsFrames = {}
local vfUpThresh = 1
local swdThresh = 0
local saInFlight = 0
local maxIns = 100
local player = ""
local vfUp = nil



-- functions
local function AddTextToFrame(frame)
	frame.text = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	frame.text:SetPoint("RIGHT", frame, -border, 0)
	frame.text:SetTextColor(1, 1, 1)
end

local function ActivateBar()
	bgFrame:Show()
	bgFrame:RegisterEvent("UNIT_POWER_FREQUENT")
end

local function DisableBar()
	bgFrame:Hide()
	bgFrame:UnregisterEvent("UNIT_POWER_FREQUENT")
end

local function VarsToPanel()
	-- apply values to widgets
	showTimesChBox:SetChecked(InsanityBarSettings.showTimes)
	showAlwaysChBox:SetChecked(InsanityBarSettings.showAlways)
	barScaleSldr:SetValue(InsanityBarSettings.barScale)
end

local function PanelToVars()
	-- get settings from widget values
	InsanityBarSettings.showTimes = showTimesChBox:GetChecked()
	InsanityBarSettings.showAlways = showAlwaysChBox:GetChecked()
	InsanityBarSettings.barScale = barScaleSldr:GetValue()
	
	-- apply some values directly for direct response
	if InsanityBarSettings.showAlways then
		ActivateBar()
	else
		DisableBar()
	end
	local widthScale = InsanityBarSettings.barScale*barWidth
	local heightScale = InsanityBarSettings.barScale*barHeight
	-- background
	bgFrame:SetSize(widthScale+2*border, heightScale+2*border)
	threshFrame:SetSize(2, heightScale+2*border)
end

local function ResetSettings()
	-- reset settings which are saved between game sessions
	-- global saved variables
	InsanityBarSettings = {showTimes=true, showAlways=false, barScale=1}
	VarsToPanel()
end

local function BuildSettingsPanel()
	-- only a convienience thing for the user
	-- options panel (esc -> Interface -> Addons)
	
	-- Add the panel to the Interface Options
	InterfaceOptions_AddCategory(addon.panel);

	-- build checkbox to show or hide drain numbers
	showTimesChBox:SetPoint("TOPLEFT", 20, -20)
	getglobal(showTimesChBox:GetName()..'Text'):SetText("Show drain times")

	-- build checkbox to show bar always
	showAlwaysChBox:SetPoint("TOPLEFT", 20, -50)
	getglobal(showAlwaysChBox:GetName()..'Text'):SetText("Show bar always")

	-- build slider for bar scale
	barScaleSldr:SetPoint("TOPLEFT", 25, -100)
	barScaleSldr:SetSize(400, 20)
	getglobal(barScaleSldr:GetName()..'Text'):SetText('Bar scale') -- sets the "title" text (top-centre of slider).
	getglobal(barScaleSldr:GetName()..'Low'):SetText('0') -- sets the left-side slider text (default is "Low").
	getglobal(barScaleSldr:GetName()..'High'):SetText('3') -- sets the right-side slider text (default is "High").
	barScaleSldr:SetMinMaxValues(0, 3)
	barScaleSldr:SetValueStep(0.1)
	barScaleSldr:SetScript("OnValueChanged", function(self, value)
		getglobal(barScaleSldr:GetName()..'Text'):SetText('Bar scale: '..math.floor(value*10)/10)
	end)
	
	VarsToPanel()

	-- function called when pressing the "default" button
	addon.panel.default = function(self)
		ResetSettings()
	end

	-- function called when pressing the "okay" button
	addon.panel.okay = function(self)
		PanelToVars()
	end
end

local function BuildBar()
	-- building the static frames for the insanity bar
	
	-- build background frame
	bgFrame:SetPoint("CENTER", UIParent)
	bgFrame.tex = bgFrame:CreateTexture("InsanityBGTex")
	bgFrame.tex:SetAllPoints(bgFrame)
	bgFrame.tex:SetColorTexture(0.1, 0.1, 0.1, 0.5)
	-- make moveable by mouse
	bgFrame:SetMovable(true)
	bgFrame:EnableMouse(true)
	bgFrame:RegisterForDrag("LeftButton")
	bgFrame:SetScript("OnDragStart", bgFrame.StartMoving)
	bgFrame:SetScript("OnDragStop", bgFrame.StopMovingOrSizing)
	bgFrame:SetUserPlaced(true)
	
	-- build bar frame
	barFrame:SetPoint("LEFT", bgFrame, "LEFT", border, 0)
	barFrame.tex = barFrame:CreateTexture("InsanityBarTex")
	barFrame.tex:SetAllPoints(barFrame)
	barFrame.tex:SetColorTexture(0.4, 0.2, 1.0, 0.7)
	-- add text frame
	AddTextToFrame(barFrame)
	
	-- build drain frame
	drainFrame:SetPoint("LEFT", bgFrame, "LEFT", border, 0)
	drainFrame.tex = drainFrame:CreateTexture("InsanityDrainTex")
	drainFrame.tex:SetAllPoints(drainFrame)
	drainFrame.tex:SetColorTexture(0.5, 0.0, 0.0, 0.5)
	drainFrame:SetFrameStrata("DIALOG")
	
	-- build cap indicator
	capFrame:SetPoint("LEFT", bgFrame, "RIGHT", -border, 0)
	capFrame.tex = capFrame:CreateTexture("InsanityCapTex")
	capFrame.tex:SetAllPoints(capFrame)
	capFrame.tex:SetColorTexture(0.2, 0.2, 0.2, 0.5)
	capFrame:SetFrameStrata("DIALOG")
	
	-- build threshold indicator
	threshFrame.tex = threshFrame:CreateTexture("InsanityThreshTex")
	threshFrame.tex:SetAllPoints(threshFrame)
	threshFrame.tex:SetColorTexture(1, 1, 1, 0.5)
end

local function UpdateInsanitySpells()
	-- dictionary of spells that grant a good amount of insanity and their corresponding amount
	-- the DoTs are not included here, but spells that could prevent dropping out of shadow form
	
	-- check for STM multiplier
	local stmMult = 1
	local _, _, _, selected = GetTalentInfo(7, 3, 1)
	if selected then
		stmMult = 3
	end
	
	-- check for RoS multiplier
	local rosMult = 1
	swdThresh = 0.2
	local _, _, _, selected = GetTalentInfo(4, 2, 1)
	if selected then
		rosMult = 3
		swdThresh = 0.35
	end
	
	-- check for insanity activation threshold
	vfUpThresh = 1
	local _, _, _, selected = GetTalentInfo(7, 1, 1)
	if selected then
		vfUpThresh = 0.7
	end
	
	spells = {}
	spells[mb] = 12 -- mindblast
	spells[vb] = 16 -- void bolt
	spells[swd] = 10*rosMult -- shadow word: death
	
	-- check for talents selected that grant insanity
	local _, _, _, selected = GetTalentInfo(1, 3, 1)
	if selected then
		spells[swv] = 25 -- shadow word: void
	end
	local _, _, _, selected = GetTalentInfo(5, 2, 1)
	if selected then
		spells[sa] = 4 -- shadowy appariations
	end
	local _, _, _, selected = GetTalentInfo(6, 2, 1)
	if selected then
		spells[sc] = 15 -- shadow crash
	end
	
	-- build extra insanity frames
	for spellId, extraIns in pairs(spells) do
		if not extraInsFrames[spellId] then
			-- make new frame because not yet in list
			local f = CreateFrame("Frame", "ExtraInsanityBarFrame"..spellId, barFrame)
			f:SetPoint("LEFT", barFrame, "RIGHT", 0, 0)
			f.tex = f:CreateTexture("ExtraInsanityBarTex"..spellId)
			f.tex:SetAllPoints(f)
			f.tex:SetColorTexture(0.7, 0.9, 1.0, 0.7)
			-- add text frame
			AddTextToFrame(f)
			
			-- also make icon
			local g = CreateFrame("Frame", "ExtraInsanityIconFrame"..spellId, f)
			g:SetPoint("LEFT", f, "LEFT", 0, 0)
			g.tex = g:CreateTexture("ExtraInsanityIconTex"..spellId)
			g.tex:SetAllPoints(g)
			local texFile = GetSpellTexture(spellId)
			g.tex:SetTexture(texFile)
			
			-- link and hide
			f.icon = g
			f:Hide()
			-- add to list
			extraInsFrames[spellId] = f
		end
	end
end

local function InsDrain(stacks)
	-- insanity drain per second based on stacks (from voidform)
	return 8+(stacks-1)/2
end

local function TimeLeft(curIns, stacks)
	-- returns the remaining time until insanity would be completely drained with stacks of voidform
	-- calculation: F(x) = INTEGRAL of InsDrain(x), y = curIns-F(x+stacks)+F(stacks), get x for y = 0
	return math.sqrt(4*curIns+stacks*stacks+30*stacks+225)-stacks-15
end

local function UpdateBar()
	-- the main routine showing current insanity and extra insanity if spells are ready
	
	local widthScale = InsanityBarSettings.barScale*barWidth
	local heightScale = InsanityBarSettings.barScale*barHeight
	local widthNorm = widthScale/maxIns
	
	-- current insanity level
	local curIns = UnitPower("player")
	barFrame:SetSize(curIns*widthNorm, heightScale)
	
	-- current drain amount
	local vfUp, _, _, counts = UnitBuff("player", vf)
	local vfStacks, timeTilDrained = 0, 0
	if vfUp then
		vfStacks = min(counts, 100)
		drainFrame:Show()
		drainFrame:SetSize(InsDrain(vfStacks)*widthNorm, heightScale)
		if InsanityBarSettings.showTimes then
			-- get seconds until insanity would be drained to 0 for current stacks
			timeTilDrained = TimeLeft(curIns, vfStacks)
			barFrame.text:Show()
			barFrame.text:SetText(string.format("%.1f", timeTilDrained))
		else
			barFrame.text:Hide()
		end
		threshFrame:Hide()
	else
		drainFrame:Hide()
		barFrame.text:Hide()
		-- voidform activate threshold
		threshFrame:Show()
		threshFrame:SetPoint("CENTER", barFrame, "LEFT", widthScale*vfUpThresh, 0)
	end
	
	-- insanity multiplier to all spells
	local globInsMult = 1
	-- surrender to madness
	local stmUp = UnitBuff("player", stm)
	if stmUp then
		globInsMult = globInsMult*3
	end
	-- power infusion
	local piUp = UnitBuff("player", pi)
	if piUp then
		globInsMult = globInsMult*1.25
	end
	
	-- append extra insanity
	local sumScaledExraIns = 0
	local prevFrame = barFrame
	for spellId, extraIns in pairs(spells) do
		-- check if spell is off cooldown
		local f = extraInsFrames[spellId]
		local start, duration = GetSpellCooldown(spellId)
		local time = GetTime()
		local cd = start + duration - time
		local cdReady = not (cd > 0 and duration > 2)
		
		-- condition for various cases
		cond = false
		if spellId == swd then
			-- special case for shadow word: death
			cond = cdReady and UnitExists("target") and UnitHealth("target")/UnitHealthMax("target") <= swdThresh
		elseif spellId == sa then
			-- special case for SA's
			cond = saInFlight > 0
			extraIns = extraIns*saInFlight
		elseif spellId == vb then
			-- special case for void bolt
			cond = cdReady and vfUp
		else
			-- normal castable spells
			cond = cdReady
		end
		
		-- show extra insanity when condition is met
		if cond then
			local scaledExtraIns = extraIns*globInsMult
			sumScaledExraIns = sumScaledExraIns+scaledExtraIns
			f:Show()
			f:SetPoint("LEFT", prevFrame, "RIGHT", 0, 0)
			f:SetSize(scaledExtraIns*widthNorm, heightScale)
			f.icon:SetSize(heightScale, heightScale)
			prevFrame = f
			-- show extra time
			if vfUp and InsanityBarSettings.showTimes then
				-- get gained seconds
				local extraTimeTilDrained = TimeLeft(scaledExtraIns+curIns, vfStacks)
				f.text:Show()
				f.text:SetText(string.format("+%.1f", extraTimeTilDrained-timeTilDrained))
			else
				f.text:Hide()
			end
		else
			f:Hide()
		end
	end
	
	-- indicate capping of extra insanity
	local totalIns = curIns+sumScaledExraIns
	if totalIns > maxIns then
		capFrame:Show()
		capFrame:SetSize((totalIns-maxIns)*widthNorm, heightScale)
	else
		capFrame:Hide()
	end
end

-- watchdog to reset SA counter in case something goes wrong
local saWatchdog = 0
mainFrame:SetScript("OnUpdate", function(self, elapsed)
	if (saWatchdog > 0) then
		saWatchdog = saWatchdog-elapsed
	else
		saInFlight = 0
	end
end)


-- events
mainFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
mainFrame:RegisterEvent("PLAYER_TALENT_UPDATE")
mainFrame:RegisterEvent("ADDON_LOADED");
mainFrame:RegisterEvent("UNIT_AURA")
mainFrame:SetScript("OnEvent", function(self, event, ...)
	-- load settings
	if event == "ADDON_LOADED" then
		if not addon.loaded then
			addon.loaded = true
			if InsanityBarSettings == nil then
				ResetSettings()
			end
			BuildSettingsPanel()
			BuildBar()
		end
	end
	
	-- look for talents selected
	if event == "PLAYER_ENTERING_WORLD" or event == "PLAYER_TALENT_UPDATE" then
		player = UnitName("player")
		UpdateInsanitySpells()
		PanelToVars()
	end
	
	-- look if voidform is up
	if event == "UNIT_AURA" then
		local unitId = ...
		if unitId == "player" then
			-- activate insanity bar while buff is up
			local buffUp = UnitBuff("player", vf)
			if buffUp then
				ActivateBar()
			else
				if not InsanityBarSettings.showAlways then
					DisableBar()
				end
			end
		end
	end
end)

bgFrame:RegisterEvent("UNIT_POWER_FREQUENT")
bgFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
bgFrame:SetScript("OnEvent", function(self, event, ...)
	-- this is fired when the main resource of a unit changes
	if event == "UNIT_POWER_FREQUENT" then
		local unit, powerType = ...
		if unit == "player" and powerType == "INSANITY" then
			UpdateBar()
		end
	end
	-- this is fired for every combat action
	if event == "COMBAT_LOG_EVENT_UNFILTERED" then
		-- we need this to track how many SA's are on the way (ideally)
		local _, spellEvent, _, _, sourceUnit, _, _, _, _, _, _, spellId, _, _, _, _, _, _, _, _, critical = ...
		-- only track stuff from us
		if sourceUnit == player then
			-- track shadowy appariations impact
			if spellId == sa and spellEvent == "SPELL_DAMAGE" then
				saInFlight = saInFlight-1
			end
			-- track crits from sw:p
			if spellId == swp and spellEvent == "SPELL_PERIODIC_DAMAGE" and critical then
				saWatchdog = 10 -- reset watchdog to x seconds
				saInFlight = saInFlight+1
			end
		end
	end
end)
